local mod = {}

local settings = require('common.settings')
local colors = settings.colors
local fonts = settings.fonts
local lap = require('sectors.lap')
local players = require('common.players')

local storage = nil
local gearbox_migration_storage = ac.storage("Gearbox_Migrated_Mph", false)
local gearbox_mph_storage = ac.storage("Gearbox_ShowMph_v2", false)
local gearbox_gal_storage = ac.storage("Gearbox_ShowGal_v2", false)
local save_settings = false
Deltabar_Modes = {
    SESSION_LAP=0,
    SESSION_OPTIMAL=1,
    FASTEST_LAP=2,
    FASTEST_OPTIMAL=3,
    PREVIOUS_LAP=4,
    MULTIPLAYER_LAP=5
}
Deltabar_ShowSectorsPerMode = table.new(table.nkeys(Deltabar_Modes), 0) -- if false visualize data on the ENTIRE lap
Gearbox_MGUKModeStartTime = -10
Leaderboard_Extended_ThisFrame = false

Sectors_UserRequestedFastestLap = false
Sectors_UserRequestedFastestLapTime = -10

Leaderboard_Multiclass_EnabledTags = table.new(20, 0)
Leaderboard_Multiclass_EnabledTime = -1
Leaderboard_Multiclass_UpdateTags = true
Leaderboard_Relative_ChangedThisFrame = false
local color_buttons = table.new(8, 0) -- TODO(cogno): save to file
local editing_color = false -- enabled when user is making is own custom color
local editing_color_idx = -1

local function append_tag(name, color)
    Leaderboard_Multiclass_EnabledTags[table.nkeys(Leaderboard_Multiclass_EnabledTags)] = {
        name=name, color=color,
    }
end

-- multiclass colors
local mc_color1 = rgbm(110 / 255, 255 / 255, 87  / 255, 1) -- green
local mc_color2 = rgbm(87  / 255, 162 / 255, 255 / 255, 1) -- blue
local mc_color3 = rgbm(255 / 255, 88  / 255, 87  / 255, 1) -- red
local mc_color4 = rgbm(233 / 255, 255 / 255, 87  / 255, 1) -- yellow
local mc_color5 = rgbm(87  / 255, 249 / 255, 255 / 255, 1) -- cyan
local mc_color6 = rgbm(255 / 255, 87  / 255, 192 / 255, 1) -- magenta
local mc_color7 = rgbm(255 / 255, 164 / 255, 87  / 255, 1) -- orange
local mc_color_count_storage = ac.storage("Multiclass_ExtraColors", 0)
local mc_tags_count_storage = ac.storage("Multiclass_TagsData", 3)


-- default as false, if no laps folder exists (or if user manually resets) we change to true and move them one by one later inside each one
CanMoveApps = {
    [APPNAMES.deltabar]    = false,
    [APPNAMES.gearbox]     = false,
    [APPNAMES.leaderboard] = false,
    [APPNAMES.map]         = false,
    [APPNAMES.real_time]   = false,
    [APPNAMES.pedals]      = false,
    [APPNAMES.radar]       = false,
    [APPNAMES.sectors]     = false,
    [APPNAMES.local_time]  = false,
    [APPNAMES.tyres]       = false,
}
local app_version = ""
function mod.init()
    storage = ac.storage{
        gearbox_scale = 1,
        gearbox_show_mph = false, -- DEPRECATED(cogno): old value, replaced by :get() and :set(...)
        gearbox_dots_window = true,
        pctime_bg = true,
        pctime_bar = false,
        maptime_bg = true,
        maptime_bar = true,
        maptime_scale = 1,
        pctime_scale = 1,
        sectors_scale = 1,
        sectors_anim_duration = 8,
        leaderboard_anim_duration = 18,
        leaderboard_scale = 1,
        leaderboard_fah = false,
        leaderboard_show_tyres = true,
        leaderboard_refresh_rate = 2000, -- in ms
        leaderboard_relative_refresh_rate = 2000, -- in ms
        leaderboard_multiclass_enabled = false,
        leaderboard_multiclass_filter_enabled = true,
        leaderboard_relative = false,
        leaderboard_people_count = 2,
        leaderboard_people_count_relative = 2,
        leaderboard_show_extended = false,
        leaderboard_show_air_temp = true,
        leaderboard_show_tarmac_temp = true,
        leaderboard_show_grip_precent = false,
        leaderboard_show_wind_info = false,
        leaderboard_session_scrolling_speed = 0,
        tyres_scale = 1,
        map_scale = 1,
        deltabar_scale = 1,
        pedals_scale = 1,
        radar_scale = 1,
        tyres_show_pedals = true,
        pedals_show_gas = true,
        pedals_show_brk = true,
        pedals_show_handbrk = false,
        pedals_show_clt = false,
        pedals_show_ffb = false,
        tyres_show_fah = false,
        tyres_show_mph = false,
        tyres_temp_colored = true,
        tyres_pressure_delta = true,
        radar_show_dots = true,
        radar_show_double_lines = true,
        radar_show_players = true,
        radar_rotate_cars = false,
        radar_extra_warning = false,
        map_low_profile = true,
        map_relative = false,
        map_remove_bg = false,
        deltabar_practice_mode = 0,
        deltabar_qualify_mode = 0,
        deltabar_race_mode = 0,
        deltabar_anim_duration = 8,
        deltabar_save_on_close = false,
        deltabar_minimized = false,
        deltabar_position = 1,
        apps_auto_scale = true,
        tyres_show_info_panel = false,
        tyres_click_priority = false,
        tyres_auto_info_panel = true,
        pedals_show_telemetry = false,
        gearbox_mguk_mode = false,
        gearbox_info_mode = true,
        gearbox_mguk_info_mode = false,
        gearbox_ers_mode = false,
        damage_panel_scale = 1,
        always_show_damage_panel = false,
        map_remove_bg_when_data_missing = false
    }
    if storage.gearbox_show_mph ~= nil and gearbox_migration_storage:get() == false then -- transition into new values
        -- NOTE(cogno): doing storage.old_value = nil will NOT delete the value, so we use an extra storage to know if we did the transition
        gearbox_mph_storage:set(storage.gearbox_show_mph)
        gearbox_gal_storage:set(storage.gearbox_show_mph)
        gearbox_migration_storage:set(true) -- transition completed
    end
    Gearbox_ShowMph = gearbox_mph_storage:get()
    Gearbox_ShowGal = gearbox_gal_storage:get()
    GearboxScale = storage.gearbox_scale
    Gearbox_DotsWindow = storage.gearbox_dots_window
    PcTime_ShowBar  = storage.pctime_bar
    PcTime_ShowBg   = storage.pctime_bg
    MapTime_ShowBar = storage.maptime_bar
    MapTime_ShowBg  = storage.maptime_bg
    MapTimeScale = storage.maptime_scale
    PcTimeScale = storage.pctime_scale
    SectorsScale = storage.sectors_scale
    Sectors_AnimDuration = storage.sectors_anim_duration
    Leaderboard_AnimDuration = storage.leaderboard_anim_duration
    LeaderboardScale = storage.leaderboard_scale
    Leaderboard_ShowFah = storage.leaderboard_fah
    Leaderboard_ShowTyres = storage.leaderboard_show_tyres
    Leaderboard_RefreshRate = storage.leaderboard_refresh_rate
    Leaderboard_Relative_RefreshRate = storage.leaderboard_relative_refresh_rate
    Leaderboard_PeopleCount = storage.leaderboard_people_count
    Leaderboard_PeopleCount_Relative = storage.leaderboard_people_count_relative
    Leaderboard_ShowExtended = storage.leaderboard_show_extended
    Leaderboard_ShowAirTemp = storage.leaderboard_show_air_temp
    Leaderboard_ShowTarmacTemp = storage.leaderboard_show_tarmac_temp
    Leaderboard_ShowGripPercent = storage.leaderboard_show_grip_precent
    Leaderboard_ShowWindInfo = storage.leaderboard_show_wind_info
    Leaderboard_SessionScrollingSpeed = storage.leaderboard_session_scrolling_speed
    TyresScale = storage.tyres_scale
    MapScale = storage.map_scale
    DeltabarScale = storage.deltabar_scale
    PedalsScale = storage.pedals_scale
    RadarScale = storage.radar_scale
    Tyres_ShowPedals = storage.tyres_show_pedals
    Pedals_ShowGas = storage.pedals_show_gas
    Pedals_ShowBrake = storage.pedals_show_brk
    Pedals_ShowHandbrake = storage.pedals_show_handbrk
    Pedals_ShowClutch = storage.pedals_show_clt
    Pedals_ShowFfb = storage.pedals_show_ffb
    Tyres_ShowFah = storage.tyres_show_fah
    Tyres_TempColored = storage.tyres_temp_colored
    Tyres_PressureDelta = storage.tyres_pressure_delta
    Tyres_ShowMph = storage.tyres_show_mph
    Radar_ShowDots = storage.radar_show_dots
    Radar_ShowDoubleLines = storage.radar_show_double_lines
    Radar_ShowPlayers = storage.radar_show_players
    Radar_RotateCars = storage.radar_rotate_cars
    Radar_MultiWarningLines = storage.radar_extra_warning
    Map_LowProfile = storage.map_low_profile
    Deltabar_PracMode = storage.deltabar_practice_mode
    Deltabar_QualMode = storage.deltabar_qualify_mode
    Deltabar_RaceMode = storage.deltabar_race_mode
    Deltabar_AnimDuration = storage.deltabar_anim_duration
    Deltabar_SaveOnClose = storage.deltabar_save_on_close
    Deltabar_Minimized = storage.deltabar_minimized
    Deltabar_Position = storage.deltabar_position
    Apps_AutoScale = storage.apps_auto_scale
    Map_Relative = storage.map_relative
    Leaderboard_Multiclass_Enabled = storage.leaderboard_multiclass_enabled
    Leaderboard_Multiclass_FilterEnabled = storage.leaderboard_multiclass_filter_enabled
    Leaderboard_Relative = storage.leaderboard_relative
    Tyres_ShowInfoPanel = storage.tyres_show_info_panel
    Tyres_ClickPriority = storage.tyres_click_priority
    Tyres_AutoInfoPanel = storage.tyres_auto_info_panel
    Pedals_ShowTelemetry = storage.pedals_show_telemetry
    Gearbox_MGUKMode =  storage.gearbox_mguk_mode
    Gearbox_InfoMode = storage.gearbox_info_mode
    Gearbox_MGUKInfoMode = storage.gearbox_mguk_info_mode
    Gearbox_ERSMode = storage.gearbox_ers_mode
    DamagePanel_Scale = storage.damage_panel_scale
    MapRemoveBG = storage.map_remove_bg
    MapRemoveBGWhenDataMissing = storage.map_remove_bg_when_data_missing
    AlwaysShowDamagePanel = storage.always_show_damage_panel

    for i=0, table.nkeys(Deltabar_Modes)-1 do
        local access_string = "deltabar_mode"..i
        local stored = ac.storage(access_string, false)
        Deltabar_ShowSectorsPerMode[i] = stored:get()
    end

    -- get the length of the tags saved by the player
    local mc_tags_count = mc_tags_count_storage:get()
    if mc_tags_count >= 1 then -- I'm scared of negative numbers, I don't know why honestly...
        -- API(cogno): find a better way instead of using the :get and :set, it looks so dumb...
        for i=0, mc_tags_count-1 do
            local s_tagname = ac.storage("Multiclass_TagsData_TagName"..i, ''):get()
            local default_color = nil
            if i % 7 == 0 then default_color = mc_color1 end
            if i % 7 == 1 then default_color = mc_color2 end
            if i % 7 == 2 then default_color = mc_color3 end
            if i % 7 == 3 then default_color = mc_color4 end
            if i % 7 == 4 then default_color = mc_color5 end
            if i % 7 == 5 then default_color = mc_color6 end
            if i % 7 == 6 then default_color = mc_color7 end
            -- TAG(cogno): ColorStorageBug
            local s_color_r = ac.storage("Multiclass_TagsData_TagColor_r"..i, "" .. default_color.r):get() -- convertion to string looks mega dumb, but it's for the stringify.parse later...
            local s_color_g = ac.storage("Multiclass_TagsData_TagColor_g"..i, "" .. default_color.g):get() -- convertion to string looks mega dumb, but it's for the stringify.parse later...
            local s_color_b = ac.storage("Multiclass_TagsData_TagColor_b"..i, "" .. default_color.b):get() -- convertion to string looks mega dumb, but it's for the stringify.parse later...
            local s_color_m = ac.storage("Multiclass_TagsData_TagColor_m"..i, "" .. default_color.mult):get() -- convertion to string looks mega dumb, but it's for the stringify.parse later...
            local s_color = rgbm(stringify.parse(s_color_r), stringify.parse(s_color_g), stringify.parse(s_color_b), stringify.parse(s_color_m))
            append_tag(s_tagname, s_color:clone())
        end
    end

    -- the player cannot edit the first 7, so we always put them here without going through save files
    color_buttons[0] = mc_color1:clone()
    color_buttons[1] = mc_color2:clone()
    color_buttons[2] = mc_color3:clone()
    color_buttons[3] = mc_color4:clone()
    color_buttons[4] = mc_color5:clone()
    color_buttons[5] = mc_color6:clone()
    color_buttons[6] = mc_color7:clone()


    -- next we get player edited colors from files
    local extra_colors = mc_color_count_storage:get()
    local current_color_buttons_length = 7
    for i=0, extra_colors-1 do
        -- TAG(cogno): ColorStorageBug
        -- TODO(cogno): explain
        local saved_color_storage_r = ac.storage("Multiclass_ExtraColor_v1_r"..i, nil)
        local saved_color_storage_g = ac.storage("Multiclass_ExtraColor_v1_g"..i, nil)
        local saved_color_storage_b = ac.storage("Multiclass_ExtraColor_v1_b"..i, nil)
        local saved_color_storage_m = ac.storage("Multiclass_ExtraColor_v1_m"..i, nil)
        local saved_r = saved_color_storage_r:get()
        local saved_g = saved_color_storage_g:get()
        local saved_b = saved_color_storage_b:get()
        local saved_m = saved_color_storage_m:get()
        if saved_r ~= nil and saved_g ~= nil and saved_b ~= nil and saved_m ~= nil  then
            color_buttons[current_color_buttons_length] = rgbm(stringify.parse(saved_r), stringify.parse(saved_g), stringify.parse(saved_b), stringify.parse(saved_m))
            current_color_buttons_length = current_color_buttons_length + 1
        end
    end

    local complete_first_time_storage = ac.storage("First_Time_Opening_Complete", false)
    local was_already_opened = complete_first_time_storage:get()
    if not was_already_opened then
        -- it's the first time opening the app EVER
        -- set that apps should be moved
        CanMoveApps[APPNAMES.deltabar]    = true
        CanMoveApps[APPNAMES.gearbox]     = true
        CanMoveApps[APPNAMES.leaderboard] = true
        CanMoveApps[APPNAMES.map]         = true
        CanMoveApps[APPNAMES.real_time]   = true
        CanMoveApps[APPNAMES.pedals]      = true
        CanMoveApps[APPNAMES.radar]       = true
        CanMoveApps[APPNAMES.sectors]     = true
        CanMoveApps[APPNAMES.local_time]  = true
        CanMoveApps[APPNAMES.tyres]       = true
        complete_first_time_storage:set(true) -- our work is done!
    end

    -- automatically gather app version so we propagate it to files
    local changelog_string = io.load("apps/lua/CMRT-Complete-HUD/CHANGELOG.txt")
    local splitted = changelog_string:split("\n")
    app_version = splitted[1]
end

local lb_full_button_toggle   = ac.ControlButton("cmrt.complete/Leaderboard Extended Toggle")
local lb_full_button_hold     = ac.ControlButton("cmrt.complete/Leaderboard Extended Hold")
local lb_button_rectract_hold = ac.ControlButton("cmrt.complete/Leaderboard Retract Hold")
local lb_button_extend        = ac.ControlButton("cmrt.complete/Leaderboard Extended")
local lb_button_rectract      = ac.ControlButton("cmrt.complete/Leaderboard Retracted")

local lb_multiclass_buttons = {
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Extended Toggle"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Extended Hold"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Retract Hold"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Open"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Close"),
}

local lb_multiclass_filter_buttons = {
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Filter Toggle"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Filter Enable Hold"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Filter Disable Hold"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Filter Enable"),
    ac.ControlButton("cmrt.complete/Leaderboard Multiclass Filter Disable"),
}

local lb_relative_buttons = {
    ac.ControlButton("cmrt.complete/Leaderboard Relative Toggle"),
    ac.ControlButton("cmrt.complete/Leaderboard Relative Enabled Hold"),
    ac.ControlButton("cmrt.complete/Leaderboard Relative Disabled Hold"),
    ac.ControlButton("cmrt.complete/Leaderboard Relative Enabled"),
    ac.ControlButton("cmrt.complete/Leaderboard Relative Disabled"),
}

local gb_full_button_toggle = ac.ControlButton("cmrt.complete/Gearbox Extended Toggle")
local gb_button_ers_hold    = ac.ControlButton("cmrt.complete/Gearbox ERS Mode Hold")
local gb_button_rpm_hold    = ac.ControlButton("cmrt.complete/Gearbox RPM Mode Hold")
local gb_button_ers_mode    = ac.ControlButton("cmrt.complete/Gearbox ERS Mode")
local gb_button_rpm_mode    = ac.ControlButton("cmrt.complete/Gearbox RPM Mode")

local map_relative_buttons = {
    ac.ControlButton("cmrt.complete/Map Display Mode"),
    ac.ControlButton("cmrt.complete/Map Relative Hold"),
    ac.ControlButton("cmrt.complete/Map Absolute Hold"),
    ac.ControlButton("cmrt.complete/Map Display Mode Relative"),
    ac.ControlButton("cmrt.complete/Map Display Mode Absolute"),
}

local tyres_button_info_panel_toggle    = ac.ControlButton("cmrt.complete/Tyres Info Panel Toggle")
local tyres_button_show_info_panel_hold = ac.ControlButton("cmrt.complete/Tyres Info Panel Hold")
local tyres_button_show_info_panel      = ac.ControlButton("cmrt.complete/Tyres Show Info Panel")
local tyres_button_hide_info_panel      = ac.ControlButton("cmrt.complete/Tyres Hide Info Panel")

local pedals_telemetry_button_toggle = ac.ControlButton("cmrt.complete/Pedals Telemetry Toggle")
local pedals_telemetry_button_hold   = ac.ControlButton("cmrt.complete/Pedals Telemetry Hold")
local pedals_telemetry_button_show   = ac.ControlButton("cmrt.complete/Pedals Telemetry Show")
local pedals_telemetry_button_hide   = ac.ControlButton("cmrt.complete/Pedals Telemetry Hide")

local sectors_button_panel_toggle    = ac.ControlButton("cmrt.complete/Sectors Panel Toggle")
local sectors_button_show_panel_hold = ac.ControlButton("cmrt.complete/Sectors Panel Hold")
local sectors_button_show_panel      = ac.ControlButton("cmrt.complete/Sectors Show Panel")
local sectors_button_hide_panel      = ac.ControlButton("cmrt.complete/Sectors Hide Panel")

local function lb_multiclass_open()
    if Leaderboard_Multiclass_Enabled == false then
        Leaderboard_Multiclass_EnabledTime = Time
    end
    Leaderboard_Multiclass_Enabled = true
    save_settings = true
end

local function lb_multiclass_close()
    if Leaderboard_Multiclass_Enabled == true then
        Leaderboard_Multiclass_EnabledTime = Time
    end
    Leaderboard_Multiclass_Enabled = false
    save_settings = true
end

local function lb_multiclass_filter_enable()
    Leaderboard_Multiclass_FilterEnabled = true
    save_settings = true
end

local function lb_multiclass_filter_disable()
    Leaderboard_Multiclass_FilterEnabled = false
    save_settings = true
end

local function lb_relative_enable()
    if Leaderboard_Relative == false then
        Leaderboard_Relative_ChangedThisFrame = true
    end
    Leaderboard_Relative = true
    save_settings = true
end

local function lb_relative_disable()
    if Leaderboard_Relative then
        Leaderboard_Relative_ChangedThisFrame = true
    end
    Leaderboard_Relative = false
    Leaderboard_Relative_ChangedThisFrame = true
    save_settings = true
end

local function map_relative_enable()
    if Map_Relative == false then
        Map_Relative_SwitchTime = Time
    end
    Map_Relative = true
    save_settings = true
end

local function map_relative_disable()
    if Map_Relative == true then
        Map_Relative_SwitchTime = Time
    end
    Map_Relative = false
    save_settings = true
end

-- button list is a list of 5 buttons, 
-- list[1] is a toggle
-- list[2] is a hold to open
-- list[3] is a hold to close
-- list[4] is to open
-- list[5] is to close
-- is_open is a bool to know if what the buttons touch is open or not
-- open_func and close_func are 2 functions that are automatically called to open or close
local function multi_button_update(button_list, is_open, open_func, close_func)
    if button_list[1]:pressed() then
        if is_open then close_func() else open_func() end
    end

    if button_list[2]:down() then open_func() end
    if button_list[2]:released() then close_func() end
    
    if button_list[3]:down() then close_func() end
    if button_list[3]:released() then open_func() end

    if button_list[4]:pressed() then open_func() end
    if button_list[5]:pressed() then close_func() end
end

function mod.update()
    -- NOTE(cogno): when you port this code into CMRT Essential remember to call mod.update() (this function) from CMRT-Essential-HUD.lua script

    -- input updates
    -- Leaderboard Extended (Toggle)
    if lb_full_button_toggle:pressed() then
        Leaderboard_ShowExtended = not Leaderboard_ShowExtended
        Leaderboard_Extended_ThisFrame = true
        Leaderboard_Extended_StartTime = Time
        save_settings = true
    end

    -- Leaderboard Extended (Hold)
    if lb_full_button_hold:down() then
        if Leaderboard_ShowExtended == false then
            Leaderboard_Extended_StartTime = Time
            Leaderboard_Extended_ThisFrame = true
        end
        Leaderboard_ShowExtended = true
        save_settings = true
    end

    if lb_full_button_hold:released() then
        Leaderboard_ShowExtended = false
        Leaderboard_Extended_ThisFrame = true
        Leaderboard_Extended_StartTime = Time
        save_settings = true
    end

    -- Leaderboard Retract (Hold)
    if lb_button_rectract_hold:down() then
        if Leaderboard_ShowExtended == true then
            Leaderboard_Extended_StartTime = Time
            Leaderboard_Extended_ThisFrame = true
        end
        Leaderboard_ShowExtended = false
        save_settings = true
    end

    if lb_button_rectract_hold:released() then
        Leaderboard_ShowExtended = true
        Leaderboard_Extended_ThisFrame = true
        Leaderboard_Extended_StartTime = Time
        save_settings = true
    end
    

    -- Leaderboard Extend
    if lb_button_extend:pressed() then
        if Leaderboard_ShowExtended == false then
            Leaderboard_Extended_StartTime = Time
        end
        Leaderboard_ShowExtended = true
        Leaderboard_Extended_ThisFrame = true
        save_settings = true
    end

    -- Leaderboard Retract
    if lb_button_rectract:pressed() then
        if Leaderboard_ShowExtended == true then
            Leaderboard_Extended_StartTime = Time
        end
        Leaderboard_ShowExtended = false
        Leaderboard_Extended_ThisFrame = true
        save_settings = true
    end

    multi_button_update(
        lb_multiclass_buttons,
        Leaderboard_Multiclass_Enabled,
        lb_multiclass_open,
        lb_multiclass_close
    )
    multi_button_update(
        lb_multiclass_filter_buttons,
        Leaderboard_Multiclass_FilterEnabled,
        lb_multiclass_filter_enable,
        lb_multiclass_filter_disable
    )
    multi_button_update(
        lb_relative_buttons,
        Leaderboard_Relative,
        lb_relative_enable,
        lb_relative_disable
    )
    

    local my_car = ac.getCar(0)
    if my_car.kersPresent then
        -- Gearbox Mode (Toggle)
        if gb_full_button_toggle:pressed() then
            if Gearbox_MGUKMode == false and Gearbox_InfoMode == false then
                Gearbox_InfoMode = true
                Gearbox_MGUKInfoMode = false
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
                Gearbox_MGUKModeStartTime = -10
            else
                Gearbox_MGUKMode = not Gearbox_MGUKMode
                Gearbox_InfoMode = not Gearbox_InfoMode
                Gearbox_MGUKInfoMode = not Gearbox_MGUKInfoMode
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
                Gearbox_MGUKModeStartTime = -10
            end
            save_settings = true
        end

        -- Gearbox Display ERS Mode
        if gb_button_ers_mode:pressed() then
            if Gearbox_MGUKMode == false and Gearbox_InfoMode == false then
                Gearbox_MGUKMode = true
                Gearbox_InfoMode = false
                Gearbox_MGUKInfoMode = true
                Gearbox_MGUKModeStartTime = -10
                save_settings = true
            else
                if Gearbox_MGUKMode == false then
                    Gearbox_ElementsOffsetTime = Time
                    Gearbox_InfoOpacityTime = Time
                    Gearbox_MGUKInfoOpacityTime = Time
                end
                Gearbox_MGUKMode = true
                Gearbox_InfoMode = false
                Gearbox_MGUKInfoMode = true
                Gearbox_MGUKModeStartTime = -10
                save_settings = true
            end
        end

        -- Gearbox Display RPM Mode
        if gb_button_rpm_mode:pressed() then
            if Gearbox_MGUKMode == false and Gearbox_InfoMode == false then
                Gearbox_InfoMode = true
                Gearbox_MGUKInfoMode = false
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
                Gearbox_MGUKModeStartTime = -10
            else
                if Gearbox_MGUKMode == true then
                    Gearbox_ElementsOffsetTime = Time
                    Gearbox_InfoOpacityTime = Time
                    Gearbox_MGUKInfoOpacityTime = Time
                end

                Gearbox_MGUKMode = false
                Gearbox_InfoMode = true
                Gearbox_MGUKInfoMode = false
                Gearbox_MGUKModeStartTime = -10
            end
            save_settings = true
        end

        -- Gearbox ERS Mode Hold
        if gb_button_ers_hold:down() then
            if Gearbox_MGUKMode == false and Gearbox_InfoMode == true then
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
            end

            Gearbox_MGUKMode = true
            Gearbox_InfoMode = false
            Gearbox_MGUKInfoMode = true
            Gearbox_MGUKModeStartTime = -10
            save_settings = true
        end

        if gb_button_ers_hold:released() then
            if Gearbox_MGUKMode == true then
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
            end

            Gearbox_MGUKMode = false
            Gearbox_InfoMode = true
            Gearbox_MGUKInfoMode = false
            Gearbox_MGUKModeStartTime = -10
            save_settings = true
        end

        -- Gearbox RPM Mode Hold
        if gb_button_rpm_hold:down() then
            if Gearbox_MGUKMode == true then
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
            end

            Gearbox_MGUKMode = false
            Gearbox_InfoMode = true
            Gearbox_MGUKInfoMode = false
            Gearbox_MGUKModeStartTime = -10
            save_settings = true
        end

        if gb_button_rpm_hold:released() then
            if Gearbox_MGUKMode == false then
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
            end

            Gearbox_MGUKMode = true
            Gearbox_InfoMode = false
            Gearbox_MGUKInfoMode = true
            Gearbox_MGUKModeStartTime = -10
            save_settings = true
        end
    end

    multi_button_update(
        map_relative_buttons,
        Map_Relative,
        map_relative_enable,
        map_relative_disable
    )

    -- Tyres
    -- Info panel (Toggle)
    if tyres_button_info_panel_toggle:pressed() then
        Tyres_ShowInfoPanel = not Tyres_ShowInfoPanel
        Tyres_ShowInfoPanelTime = Time
        Tyres_ClickPriority = true
        save_settings = true
    end


    -- Info panel (Hold)
    if tyres_button_show_info_panel_hold:down() then
        if (Tyres_ShowInfoPanel == false) then
            Tyres_ShowInfoPanelTime = Time
        end
        Tyres_ShowInfoPanel = true
        Tyres_ClickPriority = true
        save_settings = true
    end


    if tyres_button_show_info_panel_hold:released() then
        Tyres_ShowInfoPanel = false
        Tyres_ShowInfoPanelTime = Time
        Tyres_ClickPriority = false
        save_settings = true
    end

    -- Show Info panel
    if tyres_button_show_info_panel:pressed() then
        if Tyres_ShowInfoPanel == false then
            Tyres_ShowInfoPanelTime = Time
        end
        Tyres_ShowInfoPanel = true
        Tyres_ClickPriority = true
        save_settings = true
    end

    -- Hide Info panel
    if tyres_button_hide_info_panel:pressed() then
        if Tyres_ShowInfoPanel == true then
            Tyres_ShowInfoPanelTime = Time
        end
        Tyres_ShowInfoPanel = false
        Tyres_ClickPriority = false
        save_settings = true
    end

    -- Pedals
    -- Pedals Telemetry Toggle
    if pedals_telemetry_button_toggle:pressed() then
        Pedals_ShowTelemetry = not Pedals_ShowTelemetry
        Pedals_ShowTelemetryTime = Time
        save_settings = true
    end

    -- Pedals Telemetry Hold
    if pedals_telemetry_button_hold:down() then
        if (Pedals_ShowTelemetry == false) then
            Pedals_ShowTelemetryTime = Time
        end
        Pedals_ShowTelemetry = true
        save_settings = true
    end

    if pedals_telemetry_button_hold:released() then
        Pedals_ShowTelemetry = false
        Pedals_ShowTelemetryTime = Time
        save_settings = true
    end

    -- Pedals Telemetry Show
    if pedals_telemetry_button_show:pressed() then
        if Pedals_ShowTelemetry == false then
            Pedals_ShowTelemetryTime = Time
        end
        Pedals_ShowTelemetry = true
        save_settings = true
    end

    -- Pedals Telemetry Hide
    if pedals_telemetry_button_hide:pressed() then
        if Pedals_ShowTelemetry == true then
            Pedals_ShowTelemetryTime = Time
        end
        Pedals_ShowTelemetry = false
        save_settings = true
    end

    -- Sectors
    -- Fastest Lap (Toggle)
    if sectors_button_panel_toggle:pressed() then
        Sectors_UserRequestedFastestLap = not Sectors_UserRequestedFastestLap
        Sectors_UserRequestedFastestLapTime = Time
        save_settings = true
    end

    -- Fastest Lap (Hold)
    if sectors_button_show_panel_hold:down() then
        if Sectors_UserRequestedFastestLap == false then
            Sectors_UserRequestedFastestLapTime = Time
        end
        Sectors_UserRequestedFastestLap = true
        save_settings = true
    end
    if sectors_button_show_panel_hold:released() then
        Sectors_UserRequestedFastestLap = false
        Sectors_UserRequestedFastestLapTime = Time
        save_settings = true
    end

    -- Show Fastest Lap
    if sectors_button_show_panel:pressed() then
        if Sectors_UserRequestedFastestLap == false then
            Sectors_UserRequestedFastestLapTime = Time
        end
        Sectors_UserRequestedFastestLap = true
        save_settings = true
    end

    -- Hide Fastest Lap
    if sectors_button_hide_panel:pressed() then
        if Sectors_UserRequestedFastestLap == true then
            Sectors_UserRequestedFastestLapTime = Time
        end
        Sectors_UserRequestedFastestLap = false
        save_settings = true
    end
end

local function display_buttons(buttons_list, title, show_text, hide_text)
    ui.columns(1)

    ui.setNextTextBold()
    ui.newLine(-5)
    ui.text(title)
    ui.newLine(-5)
    ui.columns(2)
    ui.setColumnWidth(0, 180)

    ui.text(string.format("%s (Toggle)", title))
    ui.nextColumn()
    buttons_list[1]:control()
    ui.nextColumn()
    
    ui.text(string.format("%s (Hold)", show_text))
    ui.nextColumn()
    buttons_list[2]:control()
    ui.nextColumn()

    ui.text(string.format("%s (Hold)", hide_text))
    ui.nextColumn()
    buttons_list[3]:control()
    ui.nextColumn()

    ui.text(show_text)
    ui.nextColumn()
    buttons_list[4]:control()
    ui.nextColumn()
    
    ui.text(hide_text)
    ui.nextColumn()
    buttons_list[5]:control()
    ui.nextColumn()

    ui.columns(1)
end

local function keybinds_settings()
    ui.newLine(-5)
    ui.newLine(-5)

    
    ui.tabBar("kybinds bar", ui.TabBarFlags.FittingPolicyScroll, function()
        ui.tabItem("Leaderboard##binding_start", function()
            ui.newLine(-5)

            --
            -- Extended Leaderboard
            --
            ui.setNextTextBold()
            ui.text("Extended Leaderboard")
            ui.newLine(-5)
            ui.columns(2)
            ui.setColumnWidth(0, 180)
            
            ui.text("Full Leaderboard (Toggle)")
            ui.nextColumn()
            lb_full_button_toggle:control()
            ui.nextColumn()
            
            ui.text("Full Leaderboard (Hold)")
            ui.nextColumn()
            lb_full_button_hold:control()
            ui.nextColumn()

            ui.text("Leaderboard Retract (Hold)")
            ui.nextColumn()
            lb_button_rectract_hold:control()
            ui.nextColumn()

            ui.text("Leaderboard Extend")
            ui.nextColumn()
            lb_button_extend:control()
            ui.nextColumn()
            
            ui.text("Leaderboard Retract")
            ui.nextColumn()
            lb_button_rectract:control()
            ui.nextColumn()

            --
            -- Multiclass
            --
            display_buttons(lb_multiclass_buttons, "Multiclass", "Show Multiclass", "Hide Multiclass")
            display_buttons(lb_multiclass_filter_buttons, "Multiclass Filter", "Enable Multiclass Filter", "Disable Multiclass Filter")
            display_buttons(lb_relative_buttons, "Relative", "Show Relative", "Hide Relative")

            ui.columns(1) -- reset columns so we don't bleed into other settings
        end)
        ui.tabItem("Gearbox##binding_start", function()
            ui.newLine(-5)
            ui.columns(2)
            ui.setColumnWidth(0, 180)

              ui.text("Gearbox Mode (Toggle)")
            ui.nextColumn()
            gb_full_button_toggle:control()
            ui.nextColumn()

             ui.text("ERS Mode (Hold)")
            ui.nextColumn()
            gb_button_ers_hold:control()
            ui.nextColumn()

            ui.text("RPM Mode (Hold)")
            ui.nextColumn()
            gb_button_rpm_hold:control()
            ui.nextColumn()

            ui.text("Display ERS Mode")
            ui.nextColumn()
            gb_button_ers_mode:control()
            ui.nextColumn()

            ui.text("Display RPM Mode")
            ui.nextColumn()
            gb_button_rpm_mode:control()
            ui.nextColumn()

            ui.columns(1) -- reset columns so we don't bleed into other settings
        end)

        ui.tabItem("Map##binding_start", function()
            ui.newLine(-5)
            ui.columns(2)
            ui.setColumnWidth(0, 180)
            
            ui.text("Map Display Mode (Toggle)")
            ui.nextColumn()
            map_relative_buttons[1]:control()
            ui.nextColumn()

            ui.text("Map Relative (Hold)")
            ui.nextColumn()
            map_relative_buttons[2]:control()
            ui.nextColumn()

            ui.text("Map Absolute (Hold)")
            ui.nextColumn()
            map_relative_buttons[3]:control()
            ui.nextColumn()

            ui.text("Display Map Relative")
            ui.nextColumn()
            map_relative_buttons[4]:control()
            ui.nextColumn()

            ui.text("Display Map Absolute")
            ui.nextColumn()
            map_relative_buttons[5]:control()
            ui.nextColumn()

            ui.columns(1) -- reset columns so we don't bleed into other settings
        end)

        ui.tabItem("Tyres##binding_start", function()
            ui.newLine(-5)
            ui.columns(2)
            ui.setColumnWidth(0, 180)

            ui.text("Info panel (Toggle)")
            ui.nextColumn()
            tyres_button_info_panel_toggle:control()
            ui.nextColumn()

            ui.text("Show Info panel (Hold)")
            ui.nextColumn()
            tyres_button_show_info_panel_hold:control()
            ui.nextColumn()

            ui.text("Show Info panel")
            ui.nextColumn()
            tyres_button_show_info_panel:control()
            ui.nextColumn()

            ui.text("Hide Info panel")
            ui.nextColumn()
            tyres_button_hide_info_panel:control()
            ui.nextColumn()

            ui.columns(1) -- reset columns so we don't bleed into other settings
        end)

        ui.tabItem("Pedals##binding_start", function()
            ui.newLine(-5)
            ui.columns(2)
            ui.setColumnWidth(0, 180)

            ui.text("Telemetry (Toggle)")
            ui.nextColumn()
            pedals_telemetry_button_toggle:control()
            ui.nextColumn()

            ui.text("Telemetry (Hold)")
            ui.nextColumn()
            pedals_telemetry_button_hold:control()
            ui.nextColumn()

            ui.text("Show Telemetry")
            ui.nextColumn()
            pedals_telemetry_button_show:control()
            ui.nextColumn()

            ui.text("Hide Telemetry")
            ui.nextColumn()
            pedals_telemetry_button_hide:control()
            ui.nextColumn()

            ui.columns(1) -- reset columns so we don't bleed into other settings
        end)

        ui.tabItem("Sectors##binding_start", function()
            ui.newLine(-5)
            ui.columns(2)
            ui.setColumnWidth(0, 180)

            ui.text("Fastest Lap (Toggle)")
            ui.nextColumn()
            sectors_button_panel_toggle:control()
            ui.nextColumn()

            ui.text("Fastest Lap (Hold)")
            ui.nextColumn()
            sectors_button_show_panel_hold:control()
            ui.nextColumn()

            ui.text("Show Fastest Lap")
            ui.nextColumn()
            sectors_button_show_panel:control()
            ui.nextColumn()

            ui.text("Hide Fastest Lap")
            ui.nextColumn()
            sectors_button_hide_panel:control()
            ui.nextColumn()

            ui.columns(1) -- reset columns so we don't bleed into other settings
        end)
    end)


    ui.columns(1) -- reset columns so we don't bleed into other settings
end

local function increase_value(to_change)
    return math.clamp(to_change + 0.1, 0.5, 2)
end

local function decrease_value(to_change)
    return math.clamp(to_change - 0.1, 0.5, 2)
end



local function general_settings()
    ui.newLine(-5)
    ui.text(string.format("Change scale for every app:"))
    
    ui.newLine(-5)
    if ui.button("-", vec2(40)) then
        save_settings = true
        GearboxScale = decrease_value(GearboxScale)
        PcTimeScale  = decrease_value(PcTimeScale)
        MapTimeScale = decrease_value(MapTimeScale)
        SectorsScale = decrease_value(SectorsScale)
        LeaderboardScale = decrease_value(LeaderboardScale)
        TyresScale       = decrease_value(TyresScale)
        MapScale         = decrease_value(MapScale)
        DeltabarScale    = decrease_value(DeltabarScale)
        PedalsScale      = decrease_value(PedalsScale)
        RadarScale       = decrease_value(RadarScale)
        DamagePanel_Scale   = decrease_value(DamagePanel_Scale)
    end
    ui.sameLine()
    if ui.button(" + ", vec2(40)) then
        save_settings = true
        GearboxScale = increase_value(GearboxScale)
        PcTimeScale  = increase_value(PcTimeScale)
        MapTimeScale = increase_value(MapTimeScale)
        SectorsScale = increase_value(SectorsScale)
        LeaderboardScale = increase_value(LeaderboardScale)
        TyresScale       = increase_value(TyresScale)
        MapScale         = increase_value(MapScale)
        DeltabarScale    = increase_value(DeltabarScale)
        PedalsScale      = increase_value(PedalsScale)
        RadarScale       = increase_value(RadarScale)
        DamagePanel_Scale   = increase_value(DamagePanel_Scale)
    end
    
    ui.newLine(-5)
    if ui.button("Reset") then
        save_settings = true
        GearboxScale = 1
        PcTimeScale = 1
        MapTimeScale = 1
        SectorsScale = 1
        LeaderboardScale = 1
        TyresScale = 1
        MapScale = 1
        DeltabarScale = 1
        PedalsScale = 1
        RadarScale = 1
        DamagePanel_Scale = 1
    end
    ui.newLine(-5)
    ui.newLine(-5)

    if ui.checkbox("Scale Automatically", Apps_AutoScale) then
        Apps_AutoScale = not Apps_AutoScale
        save_settings = true
    end
    ui.newLine(-5)

    if ui.button("Reset Apps Layout") then
        -- player wants to move the apps
        CanMoveApps[APPNAMES.deltabar]      = true
        CanMoveApps[APPNAMES.gearbox]       = true
        CanMoveApps[APPNAMES.leaderboard]   = true
        CanMoveApps[APPNAMES.map]           = true
        CanMoveApps[APPNAMES.real_time]     = true
        CanMoveApps[APPNAMES.pedals]        = true
        CanMoveApps[APPNAMES.radar]         = true
        CanMoveApps[APPNAMES.sectors]       = true
        CanMoveApps[APPNAMES.damage_panel]  = true
        CanMoveApps[APPNAMES.local_time]    = true
        CanMoveApps[APPNAMES.tyres]         = true
    end

    ui.newLine(15)
    ui.drawImage(settings.get_asset("cmrt_complete_logo"), ui.getCursor(), ui.getCursor() + vec2(146, 30))
    ui.setCursor(ui.getCursor() + vec2(0, 40))
    
    ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
    ui.text(app_version)
    ui.popStyleColor()
end

local function scale_settings(old_value)
    ui.newLine(-5)
    ui.text(string.format("Change scale: %.0f%%", old_value * 100))
    
    ui.newLine(-5)
    if ui.button("-", vec2(40)) then
        old_value = old_value - 0.1
        save_settings = true
    end
    ui.sameLine()
    if ui.button(" + ", vec2(40)) then
        old_value = old_value + 0.1
        save_settings = true
    end
    
    ui.newLine(-5)
    if ui.button("Reset") then
        old_value = 1
        save_settings = true
    end
    ui.newLine(-5)
    ui.newLine(-5)
    
    old_value = math.clamp(old_value, 0.5, 2)
    return old_value
end

Gearbox_MGUKInfoOpacityTime = -10
Gearbox_InfoOpacityTime = -10
Gearbox_ElementsOffsetTime = -10
local function gearbox_settings()
    GearboxScale = scale_settings(GearboxScale)
    
    if ui.checkbox("RPM lights at limiter", Gearbox_DotsWindow) then
        Gearbox_DotsWindow = not Gearbox_DotsWindow
        save_settings = true
    end
    ui.newLine(-5)

    local my_car = ac.getCar(0)
 	if ui.checkbox("Show ERS by Default", Gearbox_ERSMode) then
        Gearbox_ERSMode = not Gearbox_ERSMode
        save_settings = true
        if my_car.kersPresent then
            if Gearbox_MGUKMode ~= Gearbox_ERSMode then
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
            end
            Gearbox_MGUKMode = Gearbox_ERSMode
            Gearbox_InfoMode = not Gearbox_ERSMode
            Gearbox_MGUKInfoMode = Gearbox_ERSMode
        else
            if Gearbox_MGUKMode == true then
                Gearbox_MGUKMode = false
                Gearbox_InfoMode = true
                Gearbox_MGUKInfoMode = false
                Gearbox_ElementsOffsetTime = Time
                Gearbox_InfoOpacityTime = Time
                Gearbox_MGUKInfoOpacityTime = Time
            end
        end
    end
    ui.newLine(-5)


    ui.text("Speed measurement units:")
    if ui.radioButton("KMH##mph", not Gearbox_ShowMph) then
        Gearbox_ShowMph = false
        save_settings = true
    end
    if ui.radioButton("MPH##mph", Gearbox_ShowMph) then
        Gearbox_ShowMph = true
        save_settings = true
    end
    ui.newLine(-5)

    ui.text("Fuel measurement units:")
    if ui.radioButton("Liters##gal", not Gearbox_ShowGal) then
        Gearbox_ShowGal = false
        save_settings = true
    end
    if ui.radioButton("Gallons##gal", Gearbox_ShowGal) then
        Gearbox_ShowGal = true
        save_settings = true
    end
end

local function maptime_settings()
    MapTimeScale = scale_settings(MapTimeScale)
    if ui.checkbox("show background", MapTime_ShowBg) then
        MapTime_ShowBg = not MapTime_ShowBg
        save_settings = true
    end
    
    if ui.checkbox("show bar", MapTime_ShowBar) then
        MapTime_ShowBar = not MapTime_ShowBar
        save_settings = true
    end
end
local function pctime_settings()
    PcTimeScale = scale_settings(PcTimeScale)
    if ui.checkbox("show background", PcTime_ShowBg) then
        PcTime_ShowBg = not PcTime_ShowBg
        save_settings = true
    end
    
    if ui.checkbox("show bar", PcTime_ShowBar) then
        PcTime_ShowBar = not PcTime_ShowBar
        save_settings = true
    end
end

local function sectors_settings()
    SectorsScale = scale_settings(SectorsScale)
    
    ui.text("Fastest Lap notification:")
    local value, changed = ui.slider("##sectors_anim_length", Sectors_AnimDuration, 5, 40, "%.0f seconds")
    Sectors_AnimDuration = value
    if changed then save_settings = true end
end

-- NOTE(cogno): table.remove(table, 0) doesn't seem to work, so we have to do this instead...
local function table_remove(to_remove_from, idx)
    -- NOTE(cogno): table.remove doesn't work with index 0, if you figure out why this becomes a single line of code
    local table_len = table.nkeys(to_remove_from)
    for i=idx, table_len-2 do
        to_remove_from[i] = to_remove_from[i+1]
    end
    to_remove_from[table_len-1] = nil
end

local function save_extra_colors()
    -- TAG(cogno): ColorStorageBug
    local extra_colors = table.nkeys(color_buttons) - 7 -- the 7 default cannot be edited nor removed
    mc_color_count_storage:set(extra_colors)
    for i=0, extra_colors-1 do
        local to_save = color_buttons[i+7]
        local saved_color_storage_r = ac.storage("Multiclass_ExtraColor_v1_r"..i, nil)
        local saved_color_storage_g = ac.storage("Multiclass_ExtraColor_v1_g"..i, nil)
        local saved_color_storage_b = ac.storage("Multiclass_ExtraColor_v1_b"..i, nil)
        local saved_color_storage_m = ac.storage("Multiclass_ExtraColor_v1_m"..i, nil)
        saved_color_storage_r:set(to_save.r)
        saved_color_storage_g:set(to_save.g)
        saved_color_storage_b:set(to_save.b)
        saved_color_storage_m:set(to_save.mult)
    end
end

local function save_multiclass_tags()
    -- TAG(cogno): ColorStorageBug
    Leaderboard_Multiclass_UpdateTags = true
    local list_len = table.nkeys(Leaderboard_Multiclass_EnabledTags)
    mc_tags_count_storage:set(list_len)
    for i=0, list_len-1 do
        local to_save = Leaderboard_Multiclass_EnabledTags[i]
        local tagname = to_save.name
        local tagcolor = to_save.color
        local saved_color_storage_r = ac.storage("Multiclass_TagsData_TagColor_r"..i, nil)
        local saved_color_storage_g = ac.storage("Multiclass_TagsData_TagColor_g"..i, nil)
        local saved_color_storage_b = ac.storage("Multiclass_TagsData_TagColor_b"..i, nil)
        local saved_color_storage_m = ac.storage("Multiclass_TagsData_TagColor_m"..i, nil)
        local saved_tag_storage = ac.storage("Multiclass_TagsData_TagName"..i, nil)
        saved_color_storage_r:set(tagcolor.r)
        saved_color_storage_g:set(tagcolor.g)
        saved_color_storage_b:set(tagcolor.b)
        saved_color_storage_m:set(tagcolor.mult)
        saved_tag_storage:set(tagname)
    end
end

Leaderboard_Extended_StartTime = -1
local popup_call_time = -1
local function leaderboard_settings()
    ui.newLine(-5)
    ui.tabBar("Tab", function()
        ui.tabItem("General", function()
            LeaderboardScale = scale_settings(LeaderboardScale)
            
            
            ui.text("Temperature Indicator:")
            if ui.radioButton("Celsius", not Leaderboard_ShowFah) then
                Leaderboard_ShowFah = false
                save_settings = true
            end
            if ui.radioButton("Fahrenheit", Leaderboard_ShowFah) then
                Leaderboard_ShowFah = true
                save_settings = true
            end
            
            ui.newLine(-5)
            if ui.checkbox("Show tyres compound", Leaderboard_ShowTyres) then
                Leaderboard_ShowTyres = not Leaderboard_ShowTyres
                save_settings = true
            end
            
            ui.newLine(-5)
            ui.text("Fastest Lap notification:")
            local value, changed = ui.slider("##sectors_anim_length", Leaderboard_AnimDuration, 5, 40, "%.0f seconds")
            Leaderboard_AnimDuration = value
            if changed then save_settings = true end
            
            ui.newLine(-5)
            ui.text("Interval refresh rate:")
            local value, changed = ui.slider("##leaderb_interv_refresh_rate", Leaderboard_RefreshRate / 1000, 0.1, 60, "%.1f s")
            Leaderboard_RefreshRate = value * 1000
            if changed then save_settings = true end

            ui.newLine(-5)
            ui.text("Relative Interval refresh rate:")
            local value, changed = ui.slider("##leaderb_relative_interv_refresh_rate", Leaderboard_Relative_RefreshRate / 1000, 0.1, 60, "%.1f s")
            Leaderboard_Relative_RefreshRate = value * 1000
            if changed then save_settings = true end
        end)
        ui.tabItem("Display", function()
            ui.newLine(-5)
            ui.newLine(-5)
            
            if ui.checkbox("Show Relative", Leaderboard_Relative) then
                Leaderboard_Relative = not Leaderboard_Relative
                Leaderboard_Relative_ChangedThisFrame = true
                save_settings = true
            end
            ui.newLine(-5)
            ui.text("Driver Count Relative:")
            local value2, change2 = ui.slider("##driver_count_relative", Leaderboard_PeopleCount_Relative, 1, 5, "%.0f people")
            Leaderboard_PeopleCount_Relative = value2
            if change2 then save_settings = true end
            ui.newLine(-5)

            if ui.checkbox("Show Extended", Leaderboard_ShowExtended) then
                Leaderboard_ShowExtended = not Leaderboard_ShowExtended
                save_settings = true
                Leaderboard_Extended_StartTime = Time
                Leaderboard_Extended_ThisFrame = true
            end
            ui.newLine(-5)

            if Leaderboard_ShowExtended then
                ui.pushDisabled()
                ui.slider("Driver count", Leaderboard_PeopleCount, 1, 5, "%.0f people")
                ui.popDisabled()
            else
                ui.text("Driver Count:")
                local value, changed = ui.slider("##driver_count", Leaderboard_PeopleCount, 1, 5, "%.0f people")
                Leaderboard_PeopleCount = value
                if changed then save_settings = true end
            end
            ui.newLine(10)
            ui.setNextTextBold()
            ui.text("Session Info:")
            if ui.checkbox("Show Air Temperature", Leaderboard_ShowAirTemp) then
                Leaderboard_ShowAirTemp = not Leaderboard_ShowAirTemp
                save_settings = true
            end

            if ui.checkbox("Show Tarmac Temperature", Leaderboard_ShowTarmacTemp) then
                Leaderboard_ShowTarmacTemp = not Leaderboard_ShowTarmacTemp
                save_settings = true
            end

            if ui.checkbox("Show Grip Percentage", Leaderboard_ShowGripPercent) then
                Leaderboard_ShowGripPercent = not Leaderboard_ShowGripPercent
                save_settings = true
            end

            if ui.checkbox("Show Wind Information", Leaderboard_ShowWindInfo) then
                Leaderboard_ShowWindInfo = not Leaderboard_ShowWindInfo
                save_settings = true
            end
            ui.newLine(-5)
            ui.text("Scrolling Speed:")
            local value, changed = ui.slider("##session_info_scrolling_speed", Leaderboard_SessionScrollingSpeed, -50, 50, "%.0f%%")
            Leaderboard_SessionScrollingSpeed = value
            if changed then save_settings = true end

        end)
        ui.tabItem("Multiclass", function()
            ui.newLine(-5)
            if ui.checkbox("Enabled", Leaderboard_Multiclass_Enabled) then
                Leaderboard_Multiclass_Enabled = not Leaderboard_Multiclass_Enabled
                save_settings = true
                Leaderboard_Multiclass_EnabledTime = Time
            end
            
            if Leaderboard_Multiclass_Enabled then
                if ui.checkbox("Show only drivers with my tag", Leaderboard_Multiclass_FilterEnabled) then
                    Leaderboard_Multiclass_FilterEnabled = not Leaderboard_Multiclass_FilterEnabled
                    save_settings = true
                end
                ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
                ui.textWrapped("Displays leaderboard data, sector data, focused drivers on the map and deltabar best times relative to your own class only.")
                ui.popStyleColor()
                ui.newLine(-5)
                ui.newLine(-5)
                
                ui.columns(4, false, "Multiclass Tags")
                ui.setColumnWidth(0, 50)
                ui.setColumnWidth(1, 200)
                ui.setColumnWidth(2, 35)
                
                local index_to_remove = -1
                local table_len = table.nkeys(Leaderboard_Multiclass_EnabledTags)
                for i = 0, table_len-1 do
                    local elem = Leaderboard_Multiclass_EnabledTags[i]
                    if elem ~= nil then
                        -- up/down
                        if ui.arrowButton(
                                'Up##' .. i,
                                ui.Direction.Up,
                                vec2(15, ui.frameHeight()),
                                i > 0 and ui.ButtonFlags.None or ui.ButtonFlags.Disabled) then
                            Leaderboard_Multiclass_EnabledTags[i] = Leaderboard_Multiclass_EnabledTags[i-1]
                            Leaderboard_Multiclass_EnabledTags[i-1] = elem
                            save_multiclass_tags()
                        end
                        ui.sameLine()
                        if ui.arrowButton(
                                'Down##' .. i,
                                ui.Direction.Down,
                                vec2(15, ui.frameHeight()),
                                i < table_len-1 and ui.ButtonFlags.None or ui.ButtonFlags.Disabled) then
                            Leaderboard_Multiclass_EnabledTags[i] = Leaderboard_Multiclass_EnabledTags[i+1]
                            Leaderboard_Multiclass_EnabledTags[i+1] = elem
                            save_multiclass_tags()
                        end
                        ui.nextColumn()

                        -- text
                        ui.setNextItemWidth(ui.availableSpaceX())
                        local classText, _, enterPressed = ui.inputText('##name' .. i, elem.name)
                        if ui.itemHovered() then
                            ui.tooltip(function()
                            ui.text("Case insensitive. Can contain multiple tags separated by comma.")
                            ui.text("Press enter to confirm changes.")
                            end)
                        end
                        if enterPressed then
                            elem.name = classText
                            save_multiclass_tags()
                        end
                        ui.nextColumn()
                        
                        -- color preview
                        ui.colorButton(
                            'Color##' .. i,
                            elem.color,
                            ui.ColorPickerFlags.None)
                        
                        
                        ui.itemPopup("popup##"..i, ui.MouseButton.Left, function()
                            popup_call_time = Time
                            ui.text("Choose color")
                            ui.newLine(-5)
                            local colors_count = table.nkeys(color_buttons)
                            for color_idx=0, colors_count-1 do
                                local color = color_buttons[color_idx]
                                if color ~= nil then
                                    if ui.colorButton("CMRT1##"..color_idx, color, ui.ColorPickerFlags.None) then
                                        elem.color = color:clone()
                                        editing_color = false
                                        save_multiclass_tags()
                                    end
                                    if color_idx >= 7 then -- we don't let you delete nor edit default colors
                                        ui.itemPopup("inner popup##" .. i .. "-"..color_idx, ui.MouseButton.Right, function()
                                            if ui.button("edit") then
                                                editing_color = true
                                                editing_color_idx = color_idx
                                                ui.closePopup()
                                                save_extra_colors()
                                            end
                                            if ui.button("delete") then
                                                ui.closePopup()
                                                table_remove(color_buttons, color_idx)
                                                editing_color = false -- deactivate so we don't get out of bounds, you can right click and edit anyway
                                                save_extra_colors()
                                            end
                                        end)
                                    end
                                    if color_idx % 7 ~= 6 then ui.sameLine() end
                                end
                            end
                            
                            if ui.iconButton(ui.Icons.Plus) then
                                editing_color = true
                                editing_color_idx = colors_count
                                color_buttons[colors_count] = colors.WHITE:clone() -- so we don't edit the original white for every app
                                save_extra_colors()
                            end
                            
                            if editing_color then
                                ui.newLine(-5)
                                local edited = ui.colorPicker("", color_buttons[editing_color_idx])
                                if edited then save_extra_colors() end
                            else
                                ui.newLine(50)
                            end

                            ui.pushAlignment(false, 1)
                            if ui.button("Ok") then
                                ui.closePopup()
                            end
                            ui.popAlignment()
                        end)
                        if Time - popup_call_time > 0.2 then editing_color = false end
                        ui.nextColumn()
    
                        -- buttons
                        if ui.iconButton(ui.Icons.Trash .. "##"..i, vec2(22, 22)) then
                            index_to_remove = i -- mark for removal (so we don't do it while iterating on it)
                        end
                        ui.nextColumn()
                    end
                end

                if index_to_remove >= 0 then
                    table_remove(Leaderboard_Multiclass_EnabledTags, index_to_remove)
                    save_multiclass_tags()
                end

                -- adding new elements
                ui.columns(3, false, "Example Id")
                ui.setColumnWidth(0, 120)
                ui.nextColumn()
                if ui.iconButton(ui.Icons.Plus .. "##mc_add", vec2(30, 20)) then
                    local new_color = nil
                    local colors_count = table.nkeys(Leaderboard_Multiclass_EnabledTags)
                    if colors_count % 7 == 0 then new_color = mc_color1:clone() end
                    if colors_count % 7 == 1 then new_color = mc_color2:clone() end
                    if colors_count % 7 == 2 then new_color = mc_color3:clone() end
                    if colors_count % 7 == 3 then new_color = mc_color4:clone() end
                    if colors_count % 7 == 4 then new_color = mc_color5:clone() end
                    if colors_count % 7 == 5 then new_color = mc_color6:clone() end
                    if colors_count % 7 == 6 then new_color = mc_color7:clone() end
                    append_tag("", new_color)
                    save_multiclass_tags()
                end
                ui.columns(1, false, "Example Id")
            end
        end)
    end)
end

Tyres_ShowInfoPanelTime = -10
local function tyres_settings()
    TyresScale = scale_settings(TyresScale)
    
    if ui.checkbox("Show attached pedals", Tyres_ShowPedals) then
        Tyres_ShowPedals = not Tyres_ShowPedals
        save_settings = true
    end
    if ui.checkbox("Show pressure delta from ideal", Tyres_PressureDelta) then
        Tyres_PressureDelta = not Tyres_PressureDelta
        save_settings = true
    end
    if ui.checkbox("Show core temperature gradient", Tyres_TempColored) then
        Tyres_TempColored = not Tyres_TempColored
        save_settings = true
    end
	if ui.checkbox("Open Info Panel automatically", Tyres_AutoInfoPanel) then
        Tyres_AutoInfoPanel = not Tyres_AutoInfoPanel
        save_settings = true
    end
    ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
    ui.textWrapped(
    "Info Panel will open automatically when you enter the pitlane and close automatically 10 seconds after you've left the pitlane.")
    ui.popStyleColor()
    ui.newLine(-5)
    ui.text("Temperature Indicator:")
    if ui.radioButton("Celsius", not Tyres_ShowFah) then
        Tyres_ShowFah = false
        save_settings = true
    end
    if ui.radioButton("Fahrenheit", Tyres_ShowFah) then
        Tyres_ShowFah = true
        save_settings = true
    end
    
    ui.newLine(-5)
    ui.text("Speed Indicator:")
    if ui.radioButton("Metric", not Tyres_ShowMph) then
        Tyres_ShowMph = false
        save_settings = true
    end
    if ui.radioButton("Imperial", Tyres_ShowMph) then
        Tyres_ShowMph = true
        save_settings = true
    end
end

Map_Relative_SwitchTime = 0

local function map_settings()
    MapScale = scale_settings(MapScale)
    
    if ui.checkbox("Low Profile Mode", Map_LowProfile) then
        Map_LowProfile = not Map_LowProfile
        save_settings = true
    end
    ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
    ui.textWrapped("In low profile mode the map will show bigger only the leader, yourself and the two positions ahead and behind you. In normal mode, instead, the map will show bigger everyone.")
    ui.popStyleColor()

    ui.newLine(-5)
    if ui.checkbox("Relative", Map_Relative) then
        Map_Relative = not Map_Relative
        Map_Relative_SwitchTime = Time
        save_settings = true
    end

    ui.newLine(-5)
    if ui.checkbox("Remove Map Background", MapRemoveBG) then
        MapRemoveBG = not MapRemoveBG
        save_settings = true
    end

    if ui.itemHovered(ui.HoveredFlags.None) then
        ui.setTooltip("Here you go, Damgam.")
    end

    ui.newLine(-5)
    if ui.checkbox("Remove Map Background When Data Is Missing", MapRemoveBGWhenDataMissing) then
        MapRemoveBGWhenDataMissing = not MapRemoveBGWhenDataMissing
        save_settings = true
    end
end

local barmode_names = {
    [0]="Session Best",
    "Session Optimal",
    "Record Best",
    "Record Optimal",
    "Previous Lap",
    "Session Record",
}

-- NOTE(cogno): be careful when you change these, because we have code that manually checks these values in deltabar.first
local barmode_choices = {
    [0]="Don't Change",
    "Session Best",
    "Session Optimal",
    "Record Best",
    "Record Optimal",
    "Previous Lap",
    "Session Record",
}

local barmode_dropdown = 0
local function deltabar_settings()
    ui.newLine(-5)
    ui.newLine(-5)
    ui.tabBar("Tab", function()
        ui.tabItem("General", function()
            DeltabarScale = scale_settings(DeltabarScale)
            
            ui.text("Info bar notification:")
            local value, changed = ui.slider("##deltabar_anim_length", Deltabar_AnimDuration, 2, 40, "%.0f seconds")
            Deltabar_AnimDuration = value
            if changed then save_settings = true end
            
            ui.newLine(-5)
            if ui.checkbox("Minimized ", Deltabar_Minimized) then
                Deltabar_Minimized = not Deltabar_Minimized
                save_settings = true
            end
            ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
            ui.textWrapped("Hides the delta bar itself, leaving only the delta on screen.")
            ui.popStyleColor()
            
            ui.newLine(-5)
            if Deltabar_Minimized then
                ui.text("Delta Position:")
                if ui.radioButton("Upper", Deltabar_Position == 0) then
                    Deltabar_Position = 0
                    save_settings = true
                end
                if ui.radioButton("Lower", Deltabar_Position == 2 or Deltabar_Position == 1) then
                    Deltabar_Position = 1
                    save_settings = true
                end
            else
                ui.text("Deltabar Position:")
                if ui.radioButton("Upper", Deltabar_Position == 0) then
                    Deltabar_Position = 0
                    save_settings = true
                end
                if ui.radioButton("Centered", Deltabar_Position == 1) then
                    Deltabar_Position = 1
                    save_settings = true
                end
                if ui.radioButton("Lower", Deltabar_Position == 2) then
                    Deltabar_Position = 2
                    save_settings = true
                end
            end

            ui.newLine(-5)
            if ui.checkbox("Save on Close", Deltabar_SaveOnClose) then
                Deltabar_SaveOnClose = not Deltabar_SaveOnClose
                save_settings = true
            end
            ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
            ui.textWrapped("Useful if notice stuttering at the end of each sector. With this setting enabled the lap/sectors data will be saved when the game closes instead of at the end of every sector.")
            ui.popStyleColor()
            ui.text("WARNING: ")
            ui.sameLine()
            ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
            ui.textWrapped("The data will be lost if the game crashes or closes unexpectedly (for example by using Alt+F4)")
            ui.popStyleColor()
        end)
        ui.tabItem("Preferences", function()
            ui.newLine(-5)
            ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
            ui.text("Choose the deltabar mode for each session")
            ui.popStyleColor()
            ui.newLine(-5)
            
            ui.text("Practice:")
            local practice_choice, prac_changed = ui.combo("##dtbar_1", Deltabar_PracMode, ui.ComboFlags.None, barmode_choices)
            Deltabar_PracMode = practice_choice
            if prac_changed then save_settings = true end

            ui.text("Qualify:")
            local qualify_choice, qual_changed = ui.combo("##dtbar_2", Deltabar_QualMode, ui.ComboFlags.None, barmode_choices)
            Deltabar_QualMode = qualify_choice
            if qual_changed then save_settings = true end
            
            ui.text("Race:")
            local race_choice, race_changed = ui.combo("##dtbar_3", Deltabar_RaceMode, ui.ComboFlags.None, barmode_choices)
            Deltabar_RaceMode = race_choice
            if race_changed then save_settings = true end
        end)
        
        ui.tabItem("Info panel", function()
            ui.newLine(-5)
            ui.pushStyleColor(ui.StyleColor.Text, colors.SETTINGS_TEXT)
            ui.text("Choose what the Info Panel shows for each mode")
            ui.popStyleColor()
            ui.newLine(-5)
            ui.text("Mode:")
            barmode_dropdown = ui.combo("##dtbar_4", barmode_dropdown, ui.ComboFlags.None, barmode_names)
            
            if ui.radioButton("Show sector", Deltabar_ShowSectorsPerMode[barmode_dropdown]) then
                Deltabar_ShowSectorsPerMode[barmode_dropdown] = true
                save_settings = true
            end
            if ui.radioButton("Show lap", not Deltabar_ShowSectorsPerMode[barmode_dropdown]) then
                Deltabar_ShowSectorsPerMode[barmode_dropdown] = false
                save_settings = true
            end
        end)
    end)
end

Pedals_ShowTelemetryTime = -10
local function pedals_settings()
    PedalsScale = scale_settings(PedalsScale)
    
    ui.text("Telemetry:")
    if ui.checkbox("Show throttle", Pedals_ShowGas) then
        Pedals_ShowGas = not Pedals_ShowGas
        save_settings = true
    end
    
    if ui.checkbox("Show brake", Pedals_ShowBrake) then
        Pedals_ShowBrake = not Pedals_ShowBrake
        save_settings = true
    end

    if ui.checkbox("Show handbrake", Pedals_ShowHandbrake) then
        Pedals_ShowHandbrake = not Pedals_ShowHandbrake
        save_settings = true
    end
    
    
    if ui.checkbox("Show clutch", Pedals_ShowClutch) then
        Pedals_ShowClutch = not Pedals_ShowClutch
        save_settings = true
    end
    
    if ui.checkbox("Show force feedback", Pedals_ShowFfb) then
        Pedals_ShowFfb = not Pedals_ShowFfb
        save_settings = true
    end
end

local function radar_settings()
    RadarScale = scale_settings(RadarScale)
    
    if ui.checkbox("Show background dots", Radar_ShowDots) then
        Radar_ShowDots = not Radar_ShowDots
        save_settings = true
    end
    if ui.checkbox("Show single horizontal line", not Radar_ShowDoubleLines) then
        Radar_ShowDoubleLines = not Radar_ShowDoubleLines
        save_settings = true
    end
    if ui.checkbox("Show other players rectangles", Radar_ShowPlayers) then
        Radar_ShowPlayers = not Radar_ShowPlayers
        save_settings = true
    end
    if Radar_ShowPlayers then
        if ui.checkbox("Rotate other players rectangles", Radar_RotateCars) then
            Radar_RotateCars = not Radar_RotateCars
            save_settings = true
        end
    end
    if ui.checkbox("Show extra warning bands", Radar_MultiWarningLines) then
        Radar_MultiWarningLines = not Radar_MultiWarningLines
        save_settings = true
    end
end

local function damage_panel_settings()
    DamagePanel_Scale = scale_settings(DamagePanel_Scale)
    if ui.checkbox("Always Show Damage Panel", AlwaysShowDamagePanel) then
        AlwaysShowDamagePanel = not AlwaysShowDamagePanel
        save_settings = true
    end
end

function mod.main()
    ui.tabBar("main bar", ui.TabBarFlags.FittingPolicyScroll, function()
        ui.tabItem("General", general_settings)
        ui.tabItem("Keybinds", keybinds_settings)
        ui.tabItem("Gearbox", gearbox_settings)
        ui.tabItem("Local time", maptime_settings)
        ui.tabItem("Real time", pctime_settings)
        ui.tabItem("Sectors", sectors_settings)
        ui.tabItem("Leaderboard", leaderboard_settings)
        ui.tabItem("Tyres", tyres_settings)
        ui.tabItem("Map", map_settings)
        ui.tabItem("Deltabar", deltabar_settings)
        ui.tabItem("Pedals", pedals_settings)
        ui.tabItem("Radar", radar_settings)
        --ui.tabItem("DamagePanel", damage_panel_settings)
        ui.tabItem("DamagePanel", damage_panel_settings)
    end)
    
    if save_settings then
        save_settings = false
        gearbox_gal_storage:set(Gearbox_ShowGal)
        gearbox_mph_storage:set(Gearbox_ShowMph)
        storage.gearbox_scale = GearboxScale
        storage.maptime_bar = MapTime_ShowBar
        storage.maptime_bg = MapTime_ShowBg
        storage.pctime_bar = PcTime_ShowBar
        storage.pctime_bg = PcTime_ShowBg
        storage.maptime_scale = MapTimeScale
        storage.pctime_scale = PcTimeScale
        storage.gearbox_dots_window = Gearbox_DotsWindow
        storage.sectors_scale = SectorsScale
        storage.sectors_anim_duration = Sectors_AnimDuration
        storage.leaderboard_anim_duration = Leaderboard_AnimDuration
        storage.leaderboard_scale = LeaderboardScale
        storage.leaderboard_fah = Leaderboard_ShowFah
        storage.leaderboard_show_tyres = Leaderboard_ShowTyres
        storage.leaderboard_refresh_rate = Leaderboard_RefreshRate
        storage.leaderboard_relative_refresh_rate = Leaderboard_Relative_RefreshRate
        storage.tyres_scale = TyresScale
        storage.map_scale = MapScale
        storage.deltabar_scale = DeltabarScale
        storage.pedals_scale = PedalsScale
        storage.radar_scale = RadarScale
        storage.tyres_show_pedals = Tyres_ShowPedals
        storage.pedals_show_gas = Pedals_ShowGas
        storage.pedals_show_brk = Pedals_ShowBrake
        storage.pedals_show_handbrk = Pedals_ShowHandbrake
        storage.pedals_show_clt = Pedals_ShowClutch
        storage.pedals_show_ffb = Pedals_ShowFfb
        storage.tyres_show_fah = Tyres_ShowFah
        storage.tyres_temp_colored = Tyres_TempColored
        storage.tyres_pressure_delta = Tyres_PressureDelta
        storage.tyres_show_mph = Tyres_ShowMph
        storage.radar_show_dots = Radar_ShowDots
        storage.radar_show_double_lines = Radar_ShowDoubleLines
        storage.radar_show_players = Radar_ShowPlayers
        storage.radar_rotate_cars = Radar_RotateCars
        storage.radar_extra_warning = Radar_MultiWarningLines
        storage.map_low_profile = Map_LowProfile
        storage.map_relative = Map_Relative
        storage.deltabar_practice_mode = Deltabar_PracMode
        storage.deltabar_qualify_mode = Deltabar_QualMode
        storage.deltabar_race_mode = Deltabar_RaceMode
        storage.deltabar_anim_duration = Deltabar_AnimDuration
        storage.deltabar_save_on_close = Deltabar_SaveOnClose
        storage.deltabar_minimized = Deltabar_Minimized
        storage.deltabar_position = Deltabar_Position
        local old_scale_value = storage.apps_auto_scale
        storage.apps_auto_scale = Apps_AutoScale
        storage.leaderboard_multiclass_enabled = Leaderboard_Multiclass_Enabled
        storage.leaderboard_multiclass_filter_enabled = Leaderboard_Multiclass_FilterEnabled
        storage.leaderboard_relative = Leaderboard_Relative
        storage.leaderboard_people_count = Leaderboard_PeopleCount
        storage.leaderboard_people_count_relative = Leaderboard_PeopleCount_Relative
        storage.leaderboard_show_extended = Leaderboard_ShowExtended
        storage.leaderboard_show_air_temp = Leaderboard_ShowAirTemp
        storage.leaderboard_show_tarmac_temp = Leaderboard_ShowTarmacTemp
        storage.leaderboard_show_grip_precent = Leaderboard_ShowGripPercent
        storage.leaderboard_show_wind_info = Leaderboard_ShowWindInfo
        storage.tyres_show_info_panel = Tyres_ShowInfoPanel 
        storage.tyres_click_priority = Tyres_ClickPriority 
        storage.pedals_show_telemetry = Pedals_ShowTelemetry 
        storage.gearbox_mguk_mode = Gearbox_MGUKMode 
        storage.gearbox_info_mode = Gearbox_InfoMode 
		storage.gearbox_ers_mode = Gearbox_ERSMode
        storage.tyres_auto_info_panel = Tyres_AutoInfoPanel
        storage.damage_panel_scale = DamagePanel_Scale
        storage.map_remove_bg = MapRemoveBG
        storage.leaderboard_session_scrolling_speed = Leaderboard_SessionScrollingSpeed
        storage.map_remove_bg_when_data_missing = MapRemoveBGWhenDataMissing
        storage.always_show_damage_panel = AlwaysShowDamagePanel

    
        for i=0, table.nkeys(Deltabar_Modes)-1 do
            local access_string = "deltabar_mode"..i
            local stored = ac.storage(access_string, false)
            stored:set(Deltabar_ShowSectorsPerMode[i])
        end

        if old_scale_value ~= Apps_AutoScale then
            -- now that we have saved if apps scale or not, we can replace the file
            if Apps_AutoScale then
                -- scale automatically
                io.copyFile("apps/lua/CMRT-Complete-HUD/auto_resize.ini", "apps/lua/CMRT-Complete-HUD/manifest.ini", false)
            else
                -- scale manually
                io.copyFile("apps/lua/CMRT-Complete-HUD/manual_resize.ini", "apps/lua/CMRT-Complete-HUD/manifest.ini", false)
            end
        end
    end
end

return mod
